! WHIZARD 1.00beta Dec 01 2000
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module mappings

  use kinds, only: default !NODEP!
  use constants, only: pi
  use parameters, only: parameter_set !NODEP!
  use parameters, only: write !NODEP!
  use particles
  use decay_trees, only: TC
  use divisions !NODEP!

  implicit none
  private

  type, public :: endpoint_mapping
     private
     logical :: set
     integer :: n_dropped, n
     real(kind=default) :: c, b, x, y, bb, a
     real(kind=default) :: chi2
  end type endpoint_mapping


  type, public :: mapping
     private
     integer :: mode
     integer :: particle_code
     real(kind=default) :: mass, width
     type(endpoint_mapping), dimension(2) :: em_msq, em_ct
  end type mapping


  public :: write
  public :: set
  public :: apply, get
  public :: map_msq_using_division, map_ct_using_division

  integer, parameter, public :: &
       & NO_MAPPING = 0, S_CHANNEL = 1, T_CHANNEL =  2, U_CHANNEL = -2
  integer, parameter, public :: ON_SHELL = -1
  type(endpoint_mapping), parameter :: NO_ENDPOINT_MAPPING = &
        & endpoint_mapping(.false., 0, 0, &
        &                  0._default, 0._default, 0._default, 0._default, &
        &                  0._default, 0._default, 0._default)
  type(mapping), parameter, public :: &
       & mapping_null = &
       & mapping(NO_MAPPING, 0, 0._default, 0._default, &
       &         (/NO_ENDPOINT_MAPPING, NO_ENDPOINT_MAPPING/), &
       &         (/NO_ENDPOINT_MAPPING, NO_ENDPOINT_MAPPING/))

  interface write
     module procedure mapping_write_unit
  end interface
  interface set
     module procedure mapping_set_code
     module procedure mapping_set_mass
  end interface
  interface apply
     module procedure mapping_apply_msq
     module procedure mapping_apply_ct
  end interface
  interface get
     module procedure mapping_get_msq
     module procedure mapping_get_ct
  end interface
  interface write
     module procedure endpoint_mapping_write_unit
     module procedure endpoint_mappings_write_unit
  end interface

contains

  subroutine mapping_write_unit(u, map, k)
    integer, intent(in) :: u
    type(mapping), intent(in) :: map
    integer(kind=TC), intent(in) :: k
    character(len=9) :: str
    select case(map%mode)
    case(S_CHANNEL); str = "s-channel"
    case(T_CHANNEL); str = "t-channel"
    case(U_CHANNEL); str = "u-channel"
    case(ON_SHELL);  str = "on-shell"
    end select
    if (map%mode /= NO_MAPPING) &
         & write(u, '(1x,A,I4,A)') &
         &   'Branch #', k, ":  Mapping (" // str // ") for particle " // &
         &   particle_name(map%particle_code)
    if (any(map%em_msq%set)) then
       write(u, '(1x,A)') "Mapping for invariant mass:"
       call write(u, map%em_msq)
    end if
    if (any(map%em_ct%set)) then
       write(u, '(1x,A)') "Mapping for cos-theta in decay:"
       call write(u, map%em_ct)
    end if
  end subroutine mapping_write_unit

  subroutine mapping_set_code(map, mode, code)
    type(mapping), intent(inout) :: map
    integer, intent(in) :: mode, code
    map%mode = mode
    map%particle_code = code
  end subroutine mapping_set_code

  subroutine mapping_set_mass(map, par, k)
    type(mapping), intent(inout) :: map
    type(parameter_set), intent(in) :: par
    integer(kind=TC), intent(in), optional :: k
    if (map%mode /= NO_MAPPING) then
       map%mass  = particle_mass (map%particle_code, par)
       if (map%mode == S_CHANNEL) then
          map%width = particle_width(map%particle_code, par)
          if (map%width <= 0) then
             call write(6, par)
             if (present(k)) call write(6, map, k)
             stop 'S-channel mapping allowed for positive width only'
          end if
       end if
    end if
  end subroutine mapping_set_mass

  subroutine mapping_apply_msq(map, r, s, msq, msq_min, msq_max, factor)
    type(mapping), intent(in) :: map
    real(kind=default), intent(in) :: r, s, msq_min, msq_max
    real(kind=default), intent(out) :: msq, factor
    real(kind=default) :: rmin, rmax, rr, msq0, tmp, x
    integer :: mode
    x = r
    factor = 1
    if (s==0) stop 'Apply mapping: zero energy'
    mode = map%mode
    msq0 = map%mass**2
    select case(mode)
    case(S_CHANNEL)
       if (msq0 < msq_min .or. msq0 > msq_max) mode = NO_MAPPING
    case(ON_SHELL)
       if (msq0 < msq_min) stop 'Mass too low for on-shell decay'
    end select
    if (any(map%em_msq%set)) then
       tmp = x
       x = endpoint_mapping_x (tmp, map%em_msq)
       factor = factor / endpoint_mapping_dy_dx (x, map%em_msq)
    end if
    select case(mode)
    case(NO_MAPPING, T_CHANNEL, U_CHANNEL)
       msq = (1-x) * msq_min + x * msq_max
       factor = factor * (msq_max - msq_min)/s
    case(S_CHANNEL)
       rmin = atan((msq_min-msq0)/(map%mass*map%width))/pi
       rmax = atan((msq_max-msq0)/(map%mass*map%width))/pi
       rr = (1-x)*rmin + x*rmax
       if (rr>-0.5_default .and. rr<.5_default) then
          tmp = tan(rr*pi)
          msq = map%mass * (map%mass + map%width * tmp)
          factor = factor &
               & * (rmax-rmin) * pi * (1 + tmp**2) * (map%mass * map%width)/s 
       else
          msq = 0
          factor = 0
       end if
    case(ON_SHELL)
       msq = msq0
    case default
       stop 'attempt to apply undefined mapping'
    end select
  end subroutine mapping_apply_msq

  subroutine mapping_get_msq(map, r, s, msq, msq_min, msq_max, factor)
    type(mapping), intent(in) :: map
    real(kind=default), intent(out) :: r
    real(kind=default), intent(in) :: s, msq, msq_min, msq_max
    real(kind=default), intent(out) :: factor
    real(kind=default) :: msq0, tmp, rmin, rmax
    integer :: mode
    factor = 1
    if (s==0) stop 'Get mapping: zero energy'
    mode = map%mode
    msq0 = map%mass**2
    select case(mode)
    case(S_CHANNEL)
       if (msq0 < msq_min .or. msq0 > msq_max)  mode = NO_MAPPING
    end select
    select case(mode)
    case(NO_MAPPING, T_CHANNEL, U_CHANNEL)
       if (msq_max/=msq_min) then
          r = (msq-msq_min) / (msq_max-msq_min)
       else
          r = 0
       end if
       factor = factor * (msq_max - msq_min)/s
    case(S_CHANNEL)
       rmin = atan((msq_min-msq0)/(map%mass*map%width))/pi
       rmax = atan((msq_max-msq0)/(map%mass*map%width))/pi
       tmp = (msq - msq0) / (map%mass * map%width)
       r = (atan(tmp)/pi - rmin) / (rmax-rmin)
       factor = factor &
            & * (rmax-rmin) * pi * (1 + tmp**2) * (map%mass * map%width) / s
    case(ON_SHELL)
       stop 'Attempt to apply multi-channel to an on-shell particle'
    case default
       stop 'Attempt to get undefined mapping'
    end select
    if (any(map%em_msq%set)) then
       tmp = r
       r = endpoint_mapping_y (tmp, map%em_msq)
       factor = factor / endpoint_mapping_dy_dx (tmp, map%em_msq)
    end if
  end subroutine mapping_get_msq

  subroutine mapping_apply_ct(map, prt, decay_p, k1, k2, sqrts, &
       &                      r, ct, st, factor)
    type(mapping), intent(in) :: map
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), dimension(:), intent(in) :: decay_p
    integer(kind=TC), intent(in) :: k1,k2
    real(kind=default), intent(in) :: sqrts, r
    real(kind=default), intent(out) :: ct, st, factor
    real(kind=default) :: x, tmp, E, p, m, eps
    integer(kind=TC) :: k
    factor = 1
    x = r
    if (any(map%em_ct%set)) then
       tmp = x
       x = endpoint_mapping_x (tmp, map%em_ct)
       factor = factor / endpoint_mapping_dy_dx (x, map%em_ct)
    end if
    select case(map%mode)
    case(NO_MAPPING, S_CHANNEL, ON_SHELL)
       tmp = 2*(1-x)
       ct = 1 - tmp
       st = sqrt(tmp*(2-tmp))
    case(T_CHANNEL, U_CHANNEL)
       select case(map%mode)
       case(T_CHANNEL); k=k2       ! used for mass and energy
       case(U_CHANNEL); k=k1
       end select
       p = decay_p(k)
       m = particle_mass(prt(k))
       E = sqrt(p**2 + m**2)
       if (E<=0) then
          ct = 1;  st = 0;  factor = 0
       else
          eps = min(2 * (map%mass**2 + m**2) / (sqrts * E), 0.4_default)
          if (eps<=0) stop 't-channel mapping fails because of zero mass'
          tmp = eps**x
          ct = (1 - 2*tmp) / (1 - 2*eps)
          if (ct>=-1 .and. ct<=1) then
             st = sqrt(1 - ct**2)
             factor = factor * (- Log(eps) * tmp / (1 - 2*eps))
          else
             ct = 1;  st = 0;  factor = 0
          end if
       end if
       if (map%mode==U_CHANNEL) ct = -ct  ! backward singularity
    case default
       stop 'attempt to apply undefined mapping'
    end select
  end subroutine mapping_apply_ct

  subroutine mapping_get_ct(map, prt, decay_p, k1, k2, sqrts, &
       &                    r, ct, factor)
    type(mapping), intent(in) :: map
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), dimension(:), intent(in) :: decay_p
    integer(kind=TC), intent(in) :: k1,k2
    real(kind=default), intent(in) :: sqrts
    real(kind=default), intent(out) :: r, factor
    real(kind=default), intent(in) :: ct
    real(kind=default) :: E, p, m, eps, leps, tmp
    integer(kind=TC) :: k
    factor = 1
    select case(map%mode)
    case(NO_MAPPING, S_CHANNEL, ON_SHELL)
       r = (ct+1)/2
    case(T_CHANNEL, U_CHANNEL)
       select case(map%mode)
       case(T_CHANNEL); k=k2
       case(U_CHANNEL); k=k1
       end select
       p = decay_p(k)
       m = particle_mass(prt(k))
       E = sqrt(p**2 + m**2)
       if (E<=0) then
          r = 0;  factor = 0
       else
          eps = min(2 * (map%mass**2 + m**2) / (sqrts * E), 0.4_default)
          if (eps<=0) stop 't-channel mapping fails because of zero mass'
          leps = Log(eps)
          select case(map%mode)
          case(T_CHANNEL);  tmp = (1-(1-2*eps)*ct)/2  ! forward singularity
          case(U_CHANNEL);  tmp = (1+(1-2*eps)*ct)/2  ! backward
          end select
          r = Log(tmp) / leps
          factor = factor * (- leps * tmp / (1-2*eps))
       end if
    case default
       stop 'attempt to get undefined mapping'
    end select
    if (any(map%em_ct%set)) then
       tmp = r
       r = endpoint_mapping_x (tmp, map%em_ct)
       factor = factor / endpoint_mapping_dy_dx (tmp, map%em_ct)
    end if
  end subroutine mapping_get_ct

  function endpoint_mapping_y (x, em) result (y)
    real(kind=default), intent(in) :: x
    type(endpoint_mapping), dimension(2), intent(in) :: em
    real(kind=default) :: y
    if (any(em%set)) then
       if (x <= 0) then
          y = 0
       else if (x < em(1)%x) then
          y = endpoint_mapping_y_s(x, em(1))
       else if (x <= 1-em(2)%x) then
          y = (em(1)%y * (1-em(2)%x - x) &
               & + (1-em(2)%y) * (x - em(1)%x)) &
               & / (1-em(2)%x - em(1)%x)
       else if (x < 1) then
          y = 1 - endpoint_mapping_y_s(1-x, em(2))
       else
          y = 1
       end if
    else
       y = x
    end if
!    print *, x, y
  end function endpoint_mapping_y

  function endpoint_mapping_y_s (x, em) result (y)
    real(kind=default), intent(in) :: x
    type(endpoint_mapping), intent(in) :: em
    real(kind=default) :: y
    if (em%set) then
       y = em%y / em%a * ((x + em%b)**(1-em%c) - em%bb)
    else
       y = x
    end if
  end function endpoint_mapping_y_s

  subroutine endpoint_mapping_write_unit(u, em)
    integer, intent(in) :: u
    type(endpoint_mapping), intent(in) :: em
    write(u, 1) "Power = ", em%c
    write(u, 1) "Offset = ", em%b
    write(u, 1) "Window (x,y) = ", em%x, em%y
    write(u, 2) "Fit: n, n0, chi2/dof", em%n_dropped, em%n, em%chi2
1   format(1x,A,3(1x,F12.5))
2   format(1x,A,2(1x,I3),1x,F12.5)
  end subroutine endpoint_mapping_write_unit

  subroutine endpoint_mappings_write_unit(u, em)
    integer, intent(in) :: u
    type(endpoint_mapping), dimension(2), intent(in) :: em
    if (em(1)%set) then
       write(u, *) "Lower end:"
       call write (u, em(1))
    end if
    if (em(2)%set) then
       write(u, *) "Upper end:"
       call write (u, em(2))
    end if
  end subroutine endpoint_mappings_write_unit

  subroutine endpoint_mapping_set_constants (em)
    type(endpoint_mapping), intent(inout) :: em
    if (em%set) then
       em%bb = em%b ** (1-em%c)
       em%a  = (em%x + em%b) ** (1-em%c) - em%bb
    else
       em%bb = 0
       em%a  = 0
    end if
  end subroutine endpoint_mapping_set_constants
  subroutine endpoint_mapping_match (em)
    type(endpoint_mapping), dimension(2), intent(inout) :: em
    real(kind=default) :: d1, d2, dd
    if (em(1)%set) d1 = (1-em(1)%c) / em(1)%a / (em(1)%x + em(1)%b)**em(1)%c
    if (em(2)%set) d2 = (1-em(2)%c) / em(2)%a / (em(2)%x + em(2)%b)**em(2)%c
    if (em(1)%set .and. em(2)%set) then
       dd = d1 + d2 + (1 - em(2)%x - em(1)%x)*d1*d2
       em(1)%y = d2 / dd
       em(2)%y = d1 / dd
    else if (em(1)%set) then
       em(1)%y = 1 / (1 + (1 - em(1)%x)*d1)
       em(2)%y = 0
    else if (em(2)%set) then
       em(1)%y = 0
       em(2)%y = 1 / (1 + (1 - em(2)%x) * d2)
    else
       em(1)%y = 0
       em(2)%y = 0
    end if
  end subroutine endpoint_mapping_match

  function endpoint_mapping_dy_dx (x, em) result (dy_dx)
    real(kind=default), intent(in) :: x
    type(endpoint_mapping), dimension(2), intent(in) :: em
    real(kind=default) :: dy_dx
    if (any(em%set)) then
       if (x <= 0) then
          dy_dx = 1
       else if (x < em(1)%x) then
          dy_dx = endpoint_mapping_dy_dx_s(x, em(1))
       else if (x <= 1-em(2)%x) then
          dy_dx = (1 - em(2)%y - em(1)%y) / (1 - em(2)%x - em(1)%x)
       else if (x < 1) then
          dy_dx = endpoint_mapping_dy_dx_s(1-x, em(2))
       else
          dy_dx = 1
       end if
    else
       dy_dx = 1
    end if
  end function endpoint_mapping_dy_dx

  function endpoint_mapping_dy_dx_s (x, em) result (dy_dx)
    real(kind=default), intent(in) :: x
    type(endpoint_mapping), intent(in) :: em
    real(kind=default) :: dy_dx
    if (em%set) then
       dy_dx = em%y * (1-em%c) / em%a / (x + em%b)**em%c
    else
       dy_dx = 1
    end if
  end function endpoint_mapping_dy_dx_s
  
  function endpoint_mapping_x (y, em) result (x)
    real(kind=default), intent(in) :: y
    type(endpoint_mapping), dimension(2), intent(in) :: em
    real(kind=default) :: x
    if (any(em%set)) then
       if (y <= 0) then
          x = 0
       else if (y < em(1)%y) then
          x = endpoint_mapping_x_s (y, em(1))
       else if (y <= 1-em(2)%y) then
          x = (em(1)%x * (1-em(2)%y - y) &
               & + (1-em(2)%x) * (y - em(1)%y)) &
               & / (1-em(2)%y - em(1)%y)
       else if (y < 1) then
          x = 1 - endpoint_mapping_x_s (1-y, em(2))
       else
          x = 1
       end if
    else
       x = y
    end if
  end function endpoint_mapping_x

  function endpoint_mapping_x_s (y, em) result (x)
    real(kind=default), intent(in) :: y
    type(endpoint_mapping), intent(in) :: em
    real(kind=default) :: x
    x = (em%a / em%y * y + em%bb) ** (1/(1-em%c)) - em%b
  end function endpoint_mapping_x_s

  subroutine map_msq_using_division (map, x)
    type(mapping), intent(inout) :: map
!    type(division), intent(inout) :: div
    real(kind=default), dimension(0:), intent(inout) :: x
!    call fit_vamp_grid_division (map%em_msq, div)
    call fit_vamp_grid_division (map%em_msq, x)
    if (any(map%em_msq%set)) then
!       call map_vamp_grid_division (div, map%em_msq)
       print *, 'Found msq fit'
       call write(6, map%em_msq)
       call map_vamp_grid_division (x, map%em_msq)
    end if
  end subroutine map_msq_using_division

  subroutine map_ct_using_division (map, x)
    type(mapping), intent(inout) :: map
!    type(division), intent(inout) :: div
    real(kind=default), dimension(0:), intent(inout) :: x
!    print *, 'trying to fit ...'
!    call fit_vamp_grid_division (map%em_ct, div)
    call fit_vamp_grid_division (map%em_ct, x)
    if (any(map%em_ct%set)) then
!       print *, '... successful'
!       call map_vamp_grid_division (div, map%em_ct)
       print *, 'Found ct fit'
       call write(6, map%em_ct)
       call map_vamp_grid_division (x, map%em_ct)
!       print *, 'Mapped grid division'
    end if
  end subroutine map_ct_using_division

  subroutine fit_vamp_grid_division (em, x_in)
    type(endpoint_mapping), dimension(2), intent(out) :: em
!    type(division), intent(in) :: div
    real(kind=default), dimension(0:), intent(in) :: x_in
!    real(kind=default), dimension(ubound(div%x,1)) :: x, dx
    real(kind=default), dimension(ubound(x_in,1)) :: x, dx
    integer :: nbin
! The DEC compiler will mess this up here if optimization>1
!    print *, 'Writing array beginning twice to check for compiler bug'
!    print *, x(0:5)
    em = NO_ENDPOINT_MAPPING
!    print *, x(0:5)
!    print * 
    nbin = ubound(x,1)
    x = (x_in(1:nbin) + x_in(0:nbin-1))/2
    dx = x_in(1:nbin) - x_in(0:nbin-1)
    call find_best_fit (x, dx, em(1))
    call find_best_fit (1 - x(nbin:1:-1), dx(nbin:1:-1), em(2))
! Print result and exit (test run)
!    print *, 'Fit results'
!    print *, em(1)
!    print *, em(2)
!    stop
!
    call endpoint_mapping_match (em)
!    if (any(em%set))  call write (6, em)
  end subroutine fit_vamp_grid_division

  subroutine map_vamp_grid_division (x, em)
!    type(division), intent(inout) :: div
    real(kind=default), dimension(0:), intent(inout) :: x
    type(endpoint_mapping), dimension(2), intent(in) :: em
    integer :: i
!    print *, 'Mapping division'
    do i=0, ubound(x,1)
!       print *, x(i), endpoint_mapping_y (x(i), em)
       x(i) = endpoint_mapping_y (x(i), em)
    end do
  end subroutine map_vamp_grid_division

  subroutine find_best_fit (x, dx, em)
    real(kind=default), dimension(:), intent(in) :: x, dx
    type(endpoint_mapping), intent(inout) :: em
    real(kind=default), parameter :: CHI2_MAX = 1._default
    real(kind=default), parameter :: CHI2_THR = 0.2_default
    integer, parameter :: N_X_WINDOWS = 8
    real(kind=default), dimension(N_X_WINDOWS), parameter :: X_WINDOW = &
         & (/ 0.45_default, 0.4_default, 0.3_default, 0.2_default, &
         &    0.1_default,  0.07_default, 0.05_default, 0.02_default /)
    real(kind=default), parameter :: chi2_fac = 0.5_default
    real(kind=default) :: chi2_save, b_save
    integer :: i
    type(endpoint_mapping) :: em_tmp
    em = NO_ENDPOINT_MAPPING
    chi2_save = 0
    b_save = 0
!    print *, 'iterating over window size'
    ITERATE_FIT: do i=1, size(X_WINDOW)
!       print *, i, X_WINDOW(i)
!       print *, 'Calling try_fit'
       call try_fit (x, dx, X_WINDOW(i), em_tmp)
!       print *, 'Done'
       if (em_tmp%set .and. em_tmp%chi2 < CHI2_MAX) then
          if (chi2_save==0 .or. em_tmp%chi2 < chi2_save) then
             chi2_save = em_tmp%chi2
             if (em_tmp%chi2 < CHI2_THR) then
                if (em_tmp%b < b_save .or. b_save==0) then
                   b_save = em_tmp%b
                else
                   exit ITERATE_FIT
                end if
             end if
             em = em_tmp
          end if
       end if
    end do ITERATE_FIT
    call endpoint_mapping_set_constants (em)
!    if (em%set) then
!       print *, 'found fit:'
!       call write(6, em)
!    else
!       print *, 'no fit found'
!    end if
!    print *, 'iteration complete'
  end subroutine find_best_fit

  subroutine try_fit (x, dx, x_win, em)
    real(kind=default), dimension(:), intent(in) :: x, dx
    real(kind=default), intent(in) :: x_win
    type(endpoint_mapping), intent(out) :: em
    integer, parameter :: MAX_BINS_DROPPED = 3
    integer, parameter :: MIN_BINS_USED = 5
    real(kind=default), parameter :: POWER_MIN = 0.5_default
    real(kind=default), parameter :: POWER_MAX = 5._default
    integer :: n, n0
    real(kind=default), dimension(:), allocatable :: xtmp, dxtmp
    em = NO_ENDPOINT_MAPPING
    do n0=1, min(MAX_BINS_DROPPED, size(dx)-1)
       if (dx(n0) <= dx(n0+1)) exit
    end do
    n = count (x(n0:size(x)-1) < x_win)
    if (n >= MIN_BINS_USED+n0-1) then
!       ! Making temporary array allocation explicit
!       print *, 'allocating x dx (', n, ')'
!       allocate (xtmp(n), dxtmp(n))
!       print *, 'done'
!       xtmp = x(n0:n+n0-1)
!       dxtmp = dx(n0:n+n0-1)
!       deallocate (xtmp, dxtmp)
!       ! Deactivated (using automatic array)
       call find_fit_solution (x(n0:n+n0-1), dx(n0:n+n0-1), em)
       em%set = em%set .and. (em%c > POWER_MIN) .and. (em%c < POWER_MAX)
       if (em%set) em%x = x_win
    end if
  end subroutine try_fit

  function fit_chi2 (x, log_dx, b, c) result (chi2)
    real(kind=default), dimension(:), intent(in) :: x, log_dx
    real(kind=default), intent(in) :: b, c
    real(kind=default) :: chi2
    real(kind=default), dimension(size(log_dx)) :: log_dx_fit
    log_dx_fit = fit_lna (x, log_dx, b, c) + c * log(x+b)
    chi2 = sum((log_dx - log_dx_fit)**2) / size(log_dx)
  end function fit_chi2

  function fit_chi2_default (log_dx) result (chi2)
    real(kind=default), dimension(:), intent(in) :: log_dx
    real(kind=default) :: chi2
    real(kind=default) :: log_dx_fit
    log_dx_fit = sum(log_dx) / size(log_dx)
    chi2 = sum((log_dx - log_dx_fit)**2) / size(log_dx)
  end function fit_chi2_default

  function fit_lna (x, log_dx, b, c) result (lna)
    real(kind=default), dimension(:), intent(in) :: x, log_dx
    real(kind=default), intent(in) :: b, c
    real(kind=default) :: lna
    lna = sum(log_dx - c * log(x+b)) / size(x)
  end function fit_lna

  subroutine fit_power (x, log_dx, b, c, dc)
    real(kind=default), dimension(:), intent(in) :: x, log_dx
    real(kind=default), intent(in) :: b
    real(kind=default), intent(out) :: c
    real(kind=default), optional, intent(out) :: dc
    real(kind=default), dimension(size(x)) :: xb, log_xb
    real(kind=default) :: sum_log_dx, sum_log_xb, sum_1_xb, c1, c2
    integer :: n
    n = size(x)
    sum_log_dx = sum(log_dx)
    xb = x + b
    log_xb = log(xb)
    sum_log_xb = sum(log_xb)
    sum_1_xb = sum(1/xb)
    c1 = (sum(log_dx * log_xb) - sum_log_dx * sum_log_xb / n) &
         & / (sum(log_xb**2) - sum_log_xb**2 / n)
    c2 = (sum(log_dx / xb) - sum_log_dx * sum_1_xb / n) &
         & / (sum(log_xb / xb) - sum_log_xb * sum_1_xb / n)
    c  = c1
    if (present(dc))  dc = c2 - c1
  end subroutine fit_power

  subroutine fit_init (b1, b2, x)
    real(kind=default), intent(out) :: b1, b2
    real(kind=default), dimension(:), intent(in) :: x
    b1 = 0
    b2 = x(size(x))
  end subroutine fit_init

  subroutine find_fit_solution (x, dx, em) 
    real(kind=default), dimension(:) :: x, dx
    type(endpoint_mapping), intent(out) :: em
    real(kind=default), dimension(size(dx)) :: log_dx
    real(kind=default) :: b1, b2, b3, c1, c2, c3, dc1, dc2, dc3
    integer, parameter :: MAX_FIT_BISECTIONS = 50
    real(kind=default), parameter :: ACCURACY_GOAL = 10*epsilon(1._default)
    integer :: i
!    print *, 'looking for fit: nbins=', size(x)
    em = NO_ENDPOINT_MAPPING
    log_dx = log(dx)
    call fit_init (b1, b2, x)
    call fit_power (x, log_dx, b1, c1, dc1)
    call fit_power (x, log_dx, b2, c2, dc2)
    if (dc1*dc2 < 0) then
       FIT_BISECTIONS: do i=1, MAX_FIT_BISECTIONS
          b3 = (dc2*b1 - dc1*b2) / (dc2 - dc1)
          call fit_power (x, log_dx, b3, c3, dc3)
          if (abs(dc3) <= ACCURACY_GOAL) then
             exit FIT_BISECTIONS
          else if (dc3*dc1 < 0) then
             b2 = b3
             dc2 = dc3
          else if (dc3*dc2 < 0) then
             b1 = b3
             dc1 = dc3
          else
             stop "Cannot happen -- something's weird in fit calculation"
          end if
       end do FIT_BISECTIONS
       em%b = b3
       em%c = c3
    else
       em%b = (b1+b2)/2
       call fit_power (x, log_dx, em%b, em%c)
    end if
    em%chi2 = fit_chi2 (x, log_dx, em%b, em%c)
    em%set = em%chi2 < fit_chi2_default (log_dx)
  end subroutine find_fit_solution


end module mappings
