! WHIZARD 1.00beta Dec 01 2000
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module whizard

  use kinds, only: default !NODEP!
  use vamp_interface, only: mc_integrate, mc_generate
  use mpi90 !NODEP!
  use limits, only: WHIZARD_ROOT
  use monte_carlo_states
  use stdhep !NODEP!
  use file_utils !NODEP!
  use histograms !NODEP!
  use events
  use vamp, only: vamp_copy_grids  !NODEP!
!  use hepevt_common
  use analyze

  implicit none
  private

  public :: whizard_integrate
  public :: whizard_generate

  logical :: integration_done = .false.
  logical :: first_event = .true.
  integer :: proc_id
  type(monte_carlo_input) :: mci
  real(kind=default) :: time_per_event
  integer, parameter :: NBIN = 20
  real(kind=default) :: weight, excess
  real(kind=default), dimension(:), allocatable :: eff, effsum
  integer :: ch, ievt, unit, uraw, uevt, chsum
  type(histogram), dimension(:), allocatable :: hh
  type(histogram)    :: h
  type(analysis_bundle) :: aa

  save

contains

  subroutine whizard_integrate(integral, error, chi2, efficiency)
    real(kind=default), intent(out), optional :: integral, error, chi2, efficiency

    if (integration_done) &
         & stop 'You must call whizard_generate before restarting the integration'
    call mpi90_init()
    call mpi90_rank (proc_id)
    if (proc_id == WHIZARD_ROOT) print *, '! WHIZARD 1.00beta (Dec 01 2000)'

    call create(mci)
    if (proc_id == WHIZARD_ROOT) call read('whizard.in', mci)
    call mpi90_broadcast(mci, WHIZARD_ROOT)

    call create(mcs, mci)

    call mc_integrate(mci, filename='whizard', time_per_event=time_per_event)

    if (proc_id == WHIZARD_ROOT) then
       if (mcs%screen_diagnostics) call write(6, mcs, mci)
       if (mcs%screen_results .and. time_per_event>0) then
          write(*, *) '!'
          write(*, '(1x,A,I10,A,I2.2,A,I2.2,1x,A)') &
               & '! Time estimate for generating 10000 unweighted events:', &
               & int(10000*time_per_event)/3600, ':', &
               & mod(int(10000*time_per_event), 3600)/60, ':', &
               & mod(int(10000*time_per_event), 60), &
               & 'hours'
       end if
  !     if (mcs%write_logfile) call write('whizard.out', mcs, mci)
    end if

    if (present(integral)) integral = mcs%integral
    if (present(error)) error = mcs%error
    if (present(chi2)) chi2 = mcs%chi2
    if (present(efficiency)) efficiency = mcs%efficiency

    integration_done = .true.
    first_event = .true.

  end subroutine whizard_integrate

  subroutine whizard_generate(single)
    logical, intent(in), optional :: single
    logical :: single_events, read_events

    if (.not.integration_done) &
         & stop 'You must call whizard_integrate before event generation'

    single_events = .false.;  if (present(single)) single_events = single

    read_events = (mcs%read_events_raw .and. .not. mcs%actual_integration_done)

    call mpi90_rank (proc_id)
    if (mcs%calls(1,5)>0 .and. proc_id==WHIZARD_ROOT) then
    allocate(hh(mcs%f%n_channels))
    call create_histogram(h, 0._default, 1._default, NBIN)
    call create(aa)
    if (proc_id == WHIZARD_ROOT) then
       call read('whizard.cut5', aa, trim(mcs%process_id))
    end if

    if (mcs%unweighted) then

       if (first_event) then
          ievt = 0
          first_event = .false.
          write(*,*) '!'
          if (mcs%integral <= 0) stop "Integral must be positive for simulation"
          mcs%luminosity = max (mcs%luminosity, mcs%calls(2,5) / mcs%integral)
          mcs%calls(2,5) = nint(mcs%integral * mcs%luminosity)
  10  format(2x,A,G11.4,A)
          call vamp_copy_grids (mcs%g, mcs%g_best)
          write(*,'(1x,A,1x,G11.4)') &
               & '! Event sample corresponds to luminosity [fb-1] =', &
               & mcs%luminosity
          write(*,*) '!'
          if (mcs%read_events_raw) then
             write(*,*) '! Looking for raw event file whizard.evx ...'
             inquire(file='whizard.evx', exist=mcs%read_events_raw)
             if (.not.mcs%read_events_raw) write(*,*) '! ... not found.  Will be generated.'
          end if
          if (mcs%read_events_raw) then
             uraw = free_unit()
             open(unit=uraw, file='whizard.evx', status='old', &
                  & action='read', form='unformatted')
             read(uraw) chsum
             if (.not.mcs%read_events_force .and. chsum /= mcs%checksum) &
                  & call checksum_error
             write(*,*) '! Reading', mcs%calls(1,5)*mcs%calls(2,5), &
                  & 'unweighted events ...'
          else
             write(*,*) '! Generating', mcs%calls(1,5)*mcs%calls(2,5), &
                  & 'unweighted events ...'
             if (mcs%write_events_raw) then
                uraw = free_unit()
                open(unit=uraw, file='whizard.evx', status='replace', &
                     & action='write', form='unformatted')
                write(uraw) mcs%checksum
             end if
          end if
          if (mcs%write_events) then
             write(*,*) '! Writing events to file whizard.evt'
             if (mcs%write_events_format == STDHEP_FMT) then
                call stdhep_init ('whizard.evt', 'WHIZARD event sample', &
                     &            mcs%calls(1,5)*mcs%calls(2,5))
             else
                uevt = free_unit()
                open(unit=uevt, file='whizard.evt', status='replace', &
                     & action='write')
             end if
          end if
          if (single_events) then
  !           call hepevt_fill(-1)
             return
          end if
       end if

       GENERATE_UNWEIGHTED: do
          ievt = ievt+1
          if (ievt > mcs%calls(1,5)*mcs%calls(2,5)) exit GENERATE_UNWEIGHTED
          if (mcs%read_events_raw) then
             mcs%evt%count = ievt
             call read(uraw, mcs%evt, fmt=RAW_FMT)
          else
             call mc_generate(excess=excess)
             if (mcs%write_events_raw) call write(uraw, mcs%evt, fmt=RAW_FMT)
          end if
          call apply(aa, mcs%evt, excess=excess)
          if (single_events) then
  !           call hepevt_fill(ievt, mcs%evt)
             return
          end if
          if (mcs%write_events) then
             call write (uevt, mcs%evt, fmt=mcs%write_events_format)
          end if
       end do GENERATE_UNWEIGHTED
  !     if (single_events) call hepevt_fill(-2)

    else

       if (single_events) &
            & stop 'Generation of single weighted events not supported'
       if (mcs%write_events) then
          write(*,*) "! *** Warning: can't write weighted events to file."
          mcs%write_events = .false.
       end if
       if (mcs%luminosity > 0) then
          write(*,*) "! *** Warning: luminosity ignored for weighted events."
       end if
       allocate(eff(mcs%f%n_channels), effsum(mcs%f%n_channels));  effsum = 0
       write(*,*) '!'
       if (mcs%read_events_raw) then
          write(*,*) '! Looking for raw event file whizard.evx ...'
          inquire(file='whizard.evx', exist=mcs%read_events_raw)
          if (.not.mcs%read_events_raw) write(*,*) '! ... not found.'
       end if
       if (mcs%read_events_raw) then
          uraw = free_unit()
          open(unit=uraw, file='whizard.evx', status='old', &
               & action='read', form='unformatted')
          read(uraw) chsum
          if (.not.mcs%read_events_force .and. chsum /= mcs%checksum) &
               & call checksum_error
          write(*,*) '! Reading', mcs%calls(1,5)*mcs%calls(2,5), &
               & 'weighted events ...'
       else
          write(*,*) '! Generating', mcs%calls(1,5)*mcs%calls(2,5), 'weighted events ...'
          if (mcs%write_events_raw) then
             uraw = free_unit()
             open(unit=uraw, file='whizard.evx', status='replace', &
                  & action='write', form='unformatted')
             write(uraw) mcs%checksum
          end if
       end if
       unit = free_unit()
       open(unit=unit, file='whizard.wgt', status='replace')
       write(unit,*) '! WHIZARD run for process ', trim(mcs%process_id)
       write(unit,*) '! Weight distribution'
       do ch=1, mcs%f%n_channels
          if (mcs%g%num_calls(ch) < 2) cycle
          write(unit,*) '!'
          write(unit,*) '! Channel #', ch 
          call create_histogram(hh(ch), 0._default, 1._default, NBIN)
          do ievt=1, int(mcs%calls(1,5)*mcs%calls(2,5) * mcs%g%weights(ch))
             if (mcs%read_events_raw) then
                mcs%evt%count = ievt
                call read(uraw, mcs%evt, fmt=RAW_FMT, &
                     &    probability=eff(ch), weight=weight, channel=ch)
             else
                call mc_generate(probability=eff(ch), weight=weight, channel=ch)
                if (mcs%write_events_raw) then
                   call write(uraw, mcs%evt, fmt=RAW_FMT, &
                        &     probability=eff(ch), weight=weight, channel=ch)
                end if
             end if
             effsum(ch) = effsum(ch) + eff(ch)
             call fill_histogram(hh(ch), eff(ch))
             call fill_histogram(h, eff(ch))
             call apply(aa, mcs%evt, weight/mcs%integral)
          end do
          call write_histogram(hh(ch), unit, over=.true.)
          write(unit,*) '!'
          write(unit,*) '! Efficiency = ', &
               & effsum(ch) / int(mcs%calls(1,5)*mcs%calls(2,5)*mcs%g%weights(ch))
       end do
       write(unit,*) '!'
       write(unit,*) '! Multi-channel:'
       call write_histogram(h, unit, over=.true.)

    end if

    mcs%it = mcs%it+1
    !call write('whizard.out', mcs%process_id, &
    !     & aa, mcs%it, mcs%integral, append=.true.)
    call write(6, aa, mcs, show_histograms=mcs%screen_histograms)
    call write('whizard-plots', TEXT_FMT, aa, mcs)
    call write('whizard-plots', GML_FMT, aa, mcs)
    if (.not.mcs%unweighted) then
       write(unit,*) '!'
       write(unit,*) '! Warmup events    = ', mcs%calls(1,4)*mcs%calls(2,4)
       write(unit,*) '! Generated events = ', mcs%calls(1,5)*mcs%calls(2,5)
       write(unit,*) '! Efficiency = ', sum(effsum) / (mcs%calls(1,5)*mcs%calls(2,5))
       close(unit)
       deallocate(eff, effsum)
    end if
    deallocate(hh)
    end if


    call mpi90_finalize()
    call destroy(mci)
    call destroy(mcs)
    if (proc_id == WHIZARD_ROOT) then
       if (mcs%read_events_raw .or. mcs%write_events_raw) close(uraw)
       if (mcs%write_events) then
          if (mcs%write_events_format == STDHEP_FMT) then
             call stdhep_end
          else
             close(uevt)
          end if
       end if
       print *, '! WHIZARD run finished.'
    end if

    integration_done = .false.

  end subroutine whizard_generate


end module whizard
